from requests.models import Response

class UnexpectedHttpResponseError(Exception):
    def __init__(self, response: Response):
        super().__init__(f'Unexpected response from {response.url} with status {response.status_code} and body "{response.text}".')



from functools import partial
import logging as log
from knowledge_mapper.auth.base_auth import BaseAuth

from knowledge_mapper.knowledge_base import (
    KnowledgeBaseRegistrationRequest,
)
from knowledge_mapper.knowledge_interaction import (
    AnswerKnowledgeInteractionRegistrationRequest,
    AskKnowledgeInteractionRegistrationRequest,
    PostKnowledgeInteractionRegistrationRequest,
    ReactKnowledgeInteractionRegistrationRequest,
)

from .data_source import DataSource
from .tke_client import TkeClient

WAIT_BEFORE_RETRY = 1


class KnowledgeMapper:
    def __init__(
        self,
        data_source: DataSource,
        authorization: BaseAuth,
        ke_url: str,
        kb_id: str,
        kb_name: str,
        kb_desc: str,
    ):
        self.data_source = data_source
        self.ke_url = ke_url
        self.kb_id = kb_id
        self.kb_name = kb_name
        self.kb_desc = kb_desc
        self.kis = dict()

        self.tke_client = TkeClient(ke_url)
        self.tke_client.connect()

        self.kb = self.tke_client.register(
            KnowledgeBaseRegistrationRequest(
                id=kb_id, name=kb_name, description=kb_desc
            )
        )
        self.data_source.set_knowledge_base(self.kb)

        self.authorization = authorization

    def start(self):
        self.kb.start_handle_loop()

    def clean_up(self):
        self.kb.unregister()

    def reregister(self):
        self.kb = self.tke_client.register(
            KnowledgeBaseRegistrationRequest(
                id=self.kb_id, name=self.kb_name, description=self.kb_desc
            )
        )

    def handle(self, ki, bindings: list[dict], requesting_kb: str):
        # For this implementation we assume that the knowledge mapper is responsible for authorisation
        # Check whether the requesting knowledge base is permitted to request the knowledge interaction
        if self.authorization is not None:
            permission = self.authorization.has_permission(requesting_kb, ki)
        else:  # no authorization is defined so, have the data source handle the request.
            permission = True

        # if permitted, then handle the request
        if permission:
            result = self.data_source.handle(ki, bindings, requesting_kb)
        else:
            result = []
        return result

    def add_knowledge_interaction(self, ki):
        if ki["type"] == "ask":
            req = AskKnowledgeInteractionRegistrationRequest(pattern=ki["pattern"])
        elif ki["type"] == "answer":
            req = AnswerKnowledgeInteractionRegistrationRequest(
                pattern=ki["pattern"], handler=partial(self.handle, ki)
            )
        elif ki["type"] == "post":
            req = PostKnowledgeInteractionRegistrationRequest(
                argument_pattern=ki["argument_pattern"],
                result_pattern=ki["result_pattern"],
            )
        elif ki["type"] == "react":
            req = ReactKnowledgeInteractionRegistrationRequest(
                argument_pattern=ki["argument_pattern"],
                result_pattern=ki["result_pattern"],
                handler=partial(self.handle, ki),
            )
        else:
            raise Exception(f"Invalid KI type: {ki['type']}")

        if "prefixes" in ki:
            req.prefixes = ki["prefixes"]

        if "name" not in ki:
            ki["name"] = None
        name = ki["name"]

        registered_ki = self.kb.register_knowledge_interaction(req, name=name)
        ki["id"] = registered_ki.id

        # The authorization object needs to be made aware of the new knowledge
        # interaction
        if self.authorization is not None:
            self.authorization.add_knowledge_interaction(ki)



def match_bindings(query: list[dict], source: list[dict]) -> list:
    matches = []
    for s in source:
        for q in query:
            q_matches = True
            for k, v in q.items():
                if not (k in s and s[k] == v):
                    q_matches = False
                    break
            if q_matches:
                matches.append(s.copy())
                break
    return matches



import logging as log

from knowledge_mapper.auth.base_auth import BaseAuth

class StaticAuth(BaseAuth):
    def __init__(self, _conf):
        pass

    def has_permission(self, kb_id, ki):
        permission = False
        if 'permitted' in ki:
            if ki['permitted'] != "*":
                # check whether the requesting kb is in the permitted list                      
                if kb_id in ki['permitted']:
                    permission = True
                else:
                    log.info('Knowledge base %s is not permitted to do this request!', kb_id)
            else: # permission is set to *, so every one is permitted
                permission = True
        else:
            permission = False
        return permission



import os
import argparse
import logging as log
import pyjson5
import sys
import importlib
import time
import signal
from knowledge_mapper.knowledge_base import KnowledgeBaseUnregistered

from knowledge_mapper.knowledge_mapper import KnowledgeMapper
from knowledge_mapper.auth.sql_auth import SqlAuth
from knowledge_mapper.auth.static_auth import StaticAuth
from knowledge_mapper.data_source import DataSource
from knowledge_mapper.sparql_source import SparqlSource
from knowledge_mapper.sql_source import SqlSource

log.basicConfig(level=log.INFO)

# This function is called when a SIGTERM signal is received. This makes it so
# that the knowledge mapper can be gracefully killed by Docker.
def handle_sigterm(*args):
    raise KeyboardInterrupt()


signal.signal(signal.SIGTERM, handle_sigterm)

DATA_SOURCE_MAX_CONNECTION_ATTEMPTS = 10
DATA_SOURCE_WAIT_BEFORE_RETRY = 3


def test_data_source(data_source: DataSource):
    success = False
    attempts = 0
    while not success:
        try:
            data_source.test()
            success = True
        except Exception as e:
            attempts += 1
            if attempts < DATA_SOURCE_MAX_CONNECTION_ATTEMPTS:
                log.warning(
                    f"Request to data source failed. Retrying in {DATA_SOURCE_WAIT_BEFORE_RETRY} s."
                )
                time.sleep(DATA_SOURCE_WAIT_BEFORE_RETRY)
            else:
                log.error(f"Request to data source failed.")
                raise e


def main():
    from . import __version__

    log.info(f"Running Knowledge Mapper {__version__}")
    parser = argparse.ArgumentParser(
        description="Expose an endpoint to a knowledge network."
    )
    parser.add_argument("config")
    args = parser.parse_args()
    with open(args.config) as config_file:
        config = pyjson5.load(config_file)

        if "sparql" in config:
            endpoint = config["sparql"]["endpoint"]
            username = None
            password = None
            if (
                "username_environment_var" in config["sparql"]
                and "password_environment_var" in config["sparql"]
            ):
                username = config["sparql"]["username_environment_var"]
                password = config["sparql"]["password_environment_var"]
            data_source = SparqlSource(endpoint, username, password)
        elif "sql_host" in config:
            data_source = SqlSource(
                config["sql_host"],
                config["sql_port"],
                config["sql_database"],
                config["sql_user"],
                config["sql_password"],
            )
        elif "plugin" in config:
            plugin_cfg = config["plugin"]
            module_name, class_name = plugin_cfg["class"].rsplit(".", 1)
            plugin_module = importlib.import_module(module_name)
            plugin_class = getattr(plugin_module, class_name)
            if "args" in plugin_cfg:
                data_source = plugin_class(*plugin_cfg["args"])
            else:
                data_source = plugin_class()
        else:
            log.error("Invalid config.")
            sys.exit(1)

        test_data_source(data_source)

        if "authorization_enabled" in config:
            if "authorization" in config:
                log.error(
                    'Cannot use both `authorization_enabled` and `authorization`, as `authorization_enabled=true` is a shorthand for `authorization={type="static"}`.'
                )
                sys.exit(1)

            if config["authorization_enabled"] == True:
                auth_config = {"type": "static"}
            elif config["authorization_enabled"] == False:
                auth_config = None
            else:
                log.error('"authorization_enabled" must be either "true" or "false"')
                sys.exit(1)
        elif "authorization" in config:
            auth_config = config["authorization"]
        else:
            auth_config = None

        if auth_config is not None:
            if auth_config["type"] == "sql":
                authorization = SqlAuth(auth_config)
            elif auth_config["type"] == "static":
                authorization = StaticAuth(auth_config)
            else:
                log.error('Unknown authorization type "%s"', auth_config["type"])
                sys.exit(1)
        else:
            authorization = None

        if "KE_ENDPOINT" in os.environ:
            ke_endpoint = os.environ.get("KE_ENDPOINT")
            if "knowledge_engine_endpoint" in config:
                log.warning(
                    'Using "KE_ENDPOINT" environment variable, but also found "knowledge_engine_endpoint" property in configuration file.'
                )
        else:
            ke_endpoint = config["knowledge_engine_endpoint"]

        km = KnowledgeMapper(
            data_source,
            authorization,
            ke_endpoint,
            config["knowledge_base"]["id"],
            config["knowledge_base"]["name"],
            config["knowledge_base"]["description"],
        )
        for ki in config["knowledge_interactions"]:
            km.add_knowledge_interaction(ki)

        try:
            while True:
                try:
                    km.start()
                    break
                except KnowledgeBaseUnregistered:
                    log.warning(
                        "Knowledge Base has been registered for some reason. Will try to reregister."
                    )
                    km.reregister()
                    for ki in config["knowledge_interactions"]:
                        km.add_knowledge_interaction(ki)
        except KeyboardInterrupt:
            log.info("Shutting down gracefully...")
        finally:
            km.clean_up()

        log.info("Goodbye.")
        exit(0)


if __name__ == "__main__":
    main()



from knowledge_mapper.utils import match_bindings

def test_match_bindings_1():
    source = [
        {"a": "a1", "b": "b1", "c": "c1"},
        {"a": "a2", "b": "b2", "c": "c2"},
        {"a": "a3", "b": "b3", "c": "c3"},
    ]
    query = [
        {"a": "a1"},
    ]

    result = match_bindings(query, source)

    expected = [
        {"a": "a1", "b": "b1", "c": "c1"},
    ]

    assert dict_list_eq(expected, result)

def test_match_bindings_2():
    source = [
        {"a": "a1", "b": "b1", "c": "c1"},
        {"a": "a2", "b": "b2", "c": "c2"},
        {"a": "a3", "b": "b3", "c": "c3"},
    ]
    query = [
        {"a": "a1"},
        {"a": "a2"},
    ]

    result = match_bindings(query, source)

    expected = [
        {"a": "a1", "b": "b1", "c": "c1"},
        {"a": "a2", "b": "b2", "c": "c2"},
    ]

    print(result)

    assert dict_list_eq(expected, result)


def dict_list_eq(l1, l2):
    sorted_l1 = sorted(sorted(d.items()) for d in l1)
    sorted_l2 = sorted(sorted(d.items()) for d in l2)
    return sorted_l1 == sorted_l2



import logging as log
import mysql.connector

from .data_source import DataSource

class SqlSource(DataSource):
    def __init__(self, host: str, port: int, database: str, user: str, password: str):
        self.host = host
        self.port = port
        self.database = database
        self.user = user
        self.password = password

    def test(self):
        log.info('Testing SQL connection.')
        self.conn = mysql.connector.connect(
            user=self.user,
            password=self.password,
            host=self.host,
            port=self.port,
            database=self.database)
        self.conn.ping()
        log.info('Succes!')


    def handle(self, ki, binding_set, requesting_kb):
        if ki['type'] == 'answer':
            return self.handle_answer(ki, binding_set, requesting_kb)
        elif ki['type'] == 'react':
            return self.handle_react(ki, binding_set, requesting_kb)


    def handle_answer(self, ki, binding_set, requesting_kb):
        sql_bindings = ()
        if binding_set:
            binding_constraints = '0 '

            # Add constraint clause with a disjunct for every binding, and in
            # each disjunct a conjunction with a conjunct for every binding
            # entry.
            for binding in binding_set:
                binding_constraints += 'OR 1 '
                for key, value in binding.items():
                    binding_constraints += f'AND {key} = %s '
                    prefix = ""
                    if key in ki['column_prefixes']:
                        prefix = ki['column_prefixes'][key]
                    # If prefixed, remove the <>'s and the prefix, otherwise,
                    # use `value` as is.
                    if prefix:
                        unprefixed = value[len(prefix) + 1:-1]
                    else:
                        unprefixed = value
                    sql_bindings += (unprefixed,)

            # HAVING is used because WHERE is evaluated before aggregations, and
            # so aliases defined in the query are unavailable.
            query = f"{ki['sql_query']} HAVING {binding_constraints}"
        else:
            query = ki['sql_query']

        # Create a cursor and execute the query.
        cursor = self.conn.cursor(named_tuple=True)
        cursor.execute(query, sql_bindings)

        result_binding_set = []
        for row in cursor:
            binding = dict()
            for variable in ki['vars']:
                # Get the value of the current variable from the current row.
                value = getattr(row, variable)

                # Check the type of the value and set the datatype accordingly.
                # TODO: Support more data types?
                if variable in ki['column_prefixes']:
                    prefix = ki['column_prefixes'][variable]
                    typed_value = f"<{prefix}{value}>"
                elif isinstance(value, int):
                    typed_value = f"\"{value}\"^^<http://www.w3.org/2001/XMLSchema#integer>"
                else:
                    # Fall back to a string literal.
                    typed_value = f"\"{value}\""

                binding[variable] = typed_value

            result_binding_set.append(binding)

        cursor.close()

        # The connection caches query results if not committed and this resulted in 
        # subsequent queries not containing the latest changes to the tables.
        self.conn.commit()


        return result_binding_set


    def handle_react(self, ki, binding_set, requesting_kb):
        if binding_set:
            for binding in binding_set:
                for statement in ki['sql_query']:
                    variables = statement['bindings']
                    sql_binding = ()
                    for variable in variables:
                        value = binding[variable]
                        # If prefixed, remove the <>'s and the prefix, otherwise,
                        # use `value` as is.
                        if variable in ki['column_prefixes']:
                            prefix = ki['column_prefixes'][variable]
                            sql_binding += (value[len(prefix) + 1:-1],)
                        elif value[0] == '"' and value[-1] == '"':
                            # This is a string literal. We want to use the
                            # string INSIDE the quotes for the SQL binding.
                            sql_binding += (value[1:-1],)
                        else:
                            # TODO: Handle other datatypes correctly if they're
                            # given with "value"^^:datatype syntax.
                            sql_binding += (value,)

                    # Create a cursor and execute the query.
                    try:
                        cursor = self.conn.cursor()
                        cursor.execute(statement['statement'], sql_binding)
                    except mysql.connector.Error as e:
                        log.warn(e)
                        self.conn.rollback()
                        continue
                # Only commit after a complete query (with multiple statements)
                # is executed.
                self.conn.commit()
            return []



import knowledge_mapper.knowledge_base as tke_kb
import knowledge_mapper.knowledge_interaction as tke_ki
import knowledge_mapper.tke_client as tke
import pytest
import asyncio
import requests
from os import environ

from uuid import uuid4

if 'KE_RUNTIME_URL' in environ:
    ke_runtime_url = environ['KE_RUNTIME_URL']
else:
    ke_runtime_url = 'http://localhost:8280/rest'

print(f'Using KE runtime at {ke_runtime_url} for tests.')

def generate_kb_id_and_name(prefix='https://example.org/'):
    random_characters = uuid4().hex
    return (f'{prefix}{random_characters}', random_characters)


@pytest.mark.asyncio
async def test_ask_answer():

    kb1_id, kb1_name = generate_kb_id_and_name()
    kb2_id, kb2_name = generate_kb_id_and_name()

    answer_ki_registered = asyncio.Event()

    # Next up are two coroutines that have to run asynchronously, because they
    # interact with eachother via the KE.

    async def kb1_task():
        client_1 = tke.TkeClient(ke_runtime_url)
        client_1.connect()
        kb1 = client_1.register(tke_kb.KnowledgeBaseRegistrationRequest(id=kb1_id, name=kb1_name, description="KB 1"))

        ki_name = 'things-that-like-other-things'
        ask_ki: tke_ki.AskKnowledgeInteraction = kb1.register_knowledge_interaction(tke_ki.AskKnowledgeInteractionRegistrationRequest(prefixes={'ex': 'http://example.org/'}, pattern='?a ex:likes ?b'), ki_name)

        # Make sure that the KIs name is in the KI's ID.
        assert(kb1.id + '/interaction/' + ki_name in requests.get(
            ke_runtime_url + '/sc/ki',
            headers = {
                'Knowledge-Base-Id': kb1.id
            }
        ).text)

        # Wait for the other KI to tell us that it has been registered.
        await answer_ki_registered.wait()

        # Trigger the knowledge interaction on the default executor.
        result = await asyncio.get_event_loop().run_in_executor(
            None,
            lambda: ask_ki.ask([])
        )

        # Assert some things
        bindings = result['bindingSet']
        assert len(bindings) == 1
        assert bindings[0]['a'] == '<han>'
        assert bindings[0]['b'] == '"coffee"'

        kb1.unregister()

    async def kb2_task():
        client_2 = tke.TkeClient(ke_runtime_url)
        client_2.connect()
        kb2 = client_2.register(tke_kb.KnowledgeBaseRegistrationRequest(id=kb2_id, name=kb2_name, description="KB 2"))

        def handler(bindings: dict, requesting_kb_id: str) -> dict:
            assert len(bindings) == 0
            assert requesting_kb_id == kb1_id
            return [{'c': '<han>', 'd': '"coffee"'}]

        kb2.register_knowledge_interaction(
            tke_ki.AnswerKnowledgeInteractionRegistrationRequest(
                prefixes={'ex': 'http://example.org/'},
                pattern='?c ex:likes ?d',
                handler=handler
            ),
        )

        # Signal that the ANSWER KI has been registered to the other task in
        # this test.
        answer_ki_registered.set()

        # Make this knowledge base's client long poll for 1 incoming knowledge
        # request.
        await asyncio.get_event_loop()\
            .run_in_executor(None, lambda: kb2.start_handle_loop(1))

        kb2.unregister()

    # Wait for both tasks to complete. (They are scheduled not sequentially, but
    # asynchronously.)
    await asyncio.gather(kb1_task(), kb2_task())

def test_reregister():
    kb_id, kb_name = generate_kb_id_and_name()
    client = tke.TkeClient(ke_runtime_url)
    client.connect()

    kb1 = client.register(tke_kb.KnowledgeBaseRegistrationRequest(id=kb_id, name=kb_name, description="KB that will be reregistered"))
    assert kb1 is not None

    # Try to register with the same id, and the reregister flag turned OFF
    kb2 = client.register(tke_kb.KnowledgeBaseRegistrationRequest(id=kb_id, name=kb_name, description="KB that will be reregistered"), reregister=False)
    assert kb2 is None

    # Try to register with the same id, and the reregister flag turned ON
    kb3 = client.register(tke_kb.KnowledgeBaseRegistrationRequest(id=kb_id, name=kb_name, description="KB that will be reregistered"), reregister=True)
    assert kb3 is not None

    # We only have to clean up the final one since it uses the same ID
    kb3.unregister()

def test_reconnect_retains_kis():
    try:
        kb_id, kb_name = generate_kb_id_and_name()
        client1 = tke.TkeClient(ke_runtime_url)
        client1.connect()

        kb1 = client1.register(tke_kb.KnowledgeBaseRegistrationRequest(id=kb_id, name=kb_name, description="KB that will be reregistered"))
        kb1.register_knowledge_interaction(tke_ki.AskKnowledgeInteractionRegistrationRequest(prefixes={'ex': 'http://example.org/'}, pattern='?a ex:likes ?b'), 'an-ask-ki')
        assert kb1 is not None

        client2 = tke.TkeClient(ke_runtime_url)
        client2.connect()
        kb1_that_reconnects = client2.get_knowledge_base(kb_id)
        assert kb1_that_reconnects.id == kb_id

        assert kb1_that_reconnects.kis == kb1.kis
        assert kb1_that_reconnects.kis_by_name == kb1.kis_by_name
    finally:
        kb1_that_reconnects.unregister()



import requests
from requests.auth import HTTPBasicAuth
import os
import logging as log
from urllib.parse import quote

from .data_source import DataSource

class SparqlSource(DataSource):
    def __init__(self, endpoint: str, env_username, env_password):
        self.endpoint = endpoint
        self.auth = False
        # env_username and env_password MUST be the names of the environment variables
        #that contain the username and password to get access to the endpoint
        # if they are present, the self.auth flag will go up
        if (env_username != None and env_username in os.environ) and (env_password != None and env_password in os.environ):
            self.auth = True
            self.username = os.environ[env_username]
            self.password = os.environ[env_password]


    def test(self):
        log.info('Testing SPARQL endpoint.')
        self.do_sparql_select('SELECT * WHERE { ?s ?p ?o . } LIMIT 0')
        log.info('Succes!')


    def handle(self, ki, binding_set, requesting_kb):
        if ki['type'] == 'answer':
            return self.handle_answer(ki, binding_set, requesting_kb)
        elif ki['type'] == 'react':
            return self.handle_react(ki, binding_set, requesting_kb)


    def handle_answer(self, ki, binding_set, requesting_kb):
        # Generate the SPARQL query based on the incoming bindings and the knowledge interaction's graph pattern.
        query = generate_sparql_select(ki, binding_set)
        # Fire the SPARQL query.
        result = self.do_sparql_select(query)
        # Restructure the bindings into TKE bindings and return it
        return restructure_bindings(result)


    def handle_react(self, ki, binding_set, requesting_kb):
        # Generate the SPARQL query based on the incoming bindings and the knowledge interaction's graph pattern.
        query = generate_sparql_insert(ki, binding_set)
        # Fire the SPARQL query.
        self.do_sparql_insert(query)
        # Restructure the bindings into TKE bindings and return it
        return []


    def do_sparql_select(self, query):
        args = {
            'headers': {
                'Accept': 'application/json'
            }
        }

        if self.auth:
            args['auth'] = HTTPBasicAuth(self.username, self.password)

        response = requests.get(
            f'{self.endpoint}?query={quote(query)}',
            **args
        )

        if response.status_code == 401:
            raise UnauthorizedError("Provide BasicAuth with system environment variables.")
        elif not response.ok:
            raise RuntimeError("Invalid response from SPARQL endpoint.  (status: {}, body: {})".format(response.status_code, response.text))

        return response.json()


    def do_sparql_insert(self, query):
        args = {
            'headers': {
                'Accept': 'application/json',
                'Content-Type': 'application/sparql-update'
            }
        }

        if self.auth:
            args['auth'] = HTTPBasicAuth(self.username, self.password)

        response = requests.post(
            f'{self.endpoint}',
            data=query,
            **args
        )

        if response.status_code == 401:
            raise UnauthorizedError("Provide BasicAuth with system environment variables with the names used in the config file.")
        elif not response.ok:
            raise RuntimeError("Invalid response from SPARQL endpoint.  (status: {}, body: {})".format(response.status_code, response.text))

        return

def generate_sparql_select(ki, incoming_bindings):
    # For all partial bindings that are actually partial, we set the
    # variables that ARE in the graph pattern, but NOT in the binding to
    # UNDEF, so that they match anything.
    for var in ki['vars']:
        for incoming_binding in incoming_bindings:
            if var not in incoming_binding:
                incoming_binding[var] = 'UNDEF'

    if 'prefixes' in ki:
        prefixes = ki['prefixes']
    else:
        prefixes = dict()

    return '''
        {prefixes_clause}
        SELECT {{variables}} WHERE {{{{
            {triple_pattern}
            {values_clause}
        }}}}
    '''.format(
            prefixes_clause='\n\t'.join(f"PREFIX {prefix}: <{ki['prefixes'][prefix]}>" for prefix in prefixes.keys()) ,
            triple_pattern=ki['pattern'],
            values_clause='''
                VALUES ({{variables}}) {{{{
                    {bindings}
                }}}}
            '''.format(
                bindings='\n'.join([
                    f'({" ".join([binding[var] for var in ki["vars"]])})' for binding in incoming_bindings
                ])
            ) if incoming_bindings else '',
        ).format(
            variables=' '.join([f'?{var}' for var in ki['vars']]),
        )


def generate_sparql_insert(ki, bindings):
    variables = ki['vars']
    if 'prefixes' in ki:
        prefixes = ki['prefixes']
    else:
        prefixes = dict()
    pref = '\n\t'.join(f"PREFIX {prefix}: <{ki['prefixes'][prefix]}>" for prefix in prefixes.keys())

    return f'''
        {pref}
        INSERT {{
            {ki['argument_pattern']}
        }} WHERE {{ VALUES ({' '.join([f'?{variable}' for variable in variables])}) {{
            {os.linesep.join(f'({" ".join([str(binding[variable]) for variable in variables])})' for binding in bindings)}
        }} }}
    '''


def restructure_bindings(sparql_results):
    restructured_binding_set = []
    for binding in sparql_results['results']['bindings']:
        restructured_binding = dict()
        for key, value in binding.items():
            if value['type'] == 'uri':
                restructured_value = f'<{value["value"]}>'
            elif binding[key]['type'] == 'literal':
                if 'datatype' in value:
                    restructured_value = f'"{value["value"]}"^^<{value["datatype"]}>'
                else:
                    restructured_value = f'"{value["value"]}"'
            restructured_binding[key] = restructured_value

        restructured_binding_set.append(restructured_binding)

    return restructured_binding_set


class UnauthorizedError(RuntimeError):
   pass
